SystemOrganization addCategory: #'AJP-Pharo-Core'!

AJPServerAdaptor subclass: #AJPPharoAdaptor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

!AJPPharoAdaptor methodsFor: 'private' stamp: 'pmm 4/8/2012 22:19'!
newServer
	^ AJPPharoServer new! !

!AJPResponseBuffer methodsFor: '*ajp-pharo-core' stamp: 'pmm 2/26/2012 17:41'!
basicNextPut: anObject
	self nextPut: anObject! !

!AJPResponseBuffer methodsFor: '*ajp-pharo-core' stamp: 'pmm 8/13/2012 21:31'!
basicNextPutAll: anObject
	self nextPutAll: anObject! !

!AJPResponseBuffer methodsFor: '*ajp-pharo-core' stamp: 'pmm 4/6/2012 19:27'!
ensureCapacity: anInteger
	| bufferSize |
	bufferSize := array size.
	(packetIndex bitOr: (packetNumber - 1 bitShift: 13)) + anInteger + 8 "may need 7 byte header and 1 byte trailer" > bufferSize ifTrue: [
		| newSize newArray oldPosition |
		
		"build new array"
		"should factor in overhead for header and trailer"
		newSize := (bufferSize + anInteger) * 2.
		newArray := array species new: newSize.
		newArray replaceFrom: 1 to: bufferSize with: array startingAt: 1.
		
		"use new array"
		array := newArray.
		oldPosition := fullBuffer position.
		fullBuffer array: newArray.
		fullBuffer position: oldPosition ]! !

!AJPResponseBuffer methodsFor: '*ajp-pharo-core' stamp: 'pmm 5/1/2012 14:39'!
greaseNext: anInteger putAll: aCollection startingAt: startIndex
	| firstPacketChunk endIndex |
	
	"checking"
	anInteger = 0
		ifTrue: [ ^ self ].
	self prepareWrite: anInteger.
	
	"writing"
	firstPacketChunk :=  anInteger min: 16r2000 - packetIndex. "self maximumPacketSize (because packetIndex) - 1 (trailer)"
	"first chunk"
	endIndex := firstPacketChunk + startIndex - 1.
	self writeDirect: aCollection from: startIndex to: endIndex.
	"middle chunks"
	endIndex + 1 to: anInteger - 16r1FF8 + startIndex  by: 16r1FF8 "self maximumChunkSize" do: [ :chunkStart |
		self finishPacketIfAtEnd.
		endIndex := chunkStart + 16r1FF7. "self maximumChunkSize - 1"
		self writeDirect: aCollection from: chunkStart to: endIndex ].
	"last chunk (may not exist)"
	anInteger > (endIndex - startIndex + 1)  ifTrue: [
		self finishPacketIfAtEnd.
		self writeDirect: aCollection from: endIndex + 1 to: anInteger + startIndex - 1 ]! !

!AJPResponseBuffer methodsFor: '*ajp-pharo-core' stamp: 'pmm 2/27/2012 19:46'!
isBinary
	"because of UTF8TextConverter >> #nextPut:toStream:"
	^ false! !

AJPServer subclass: #AJPPharoServer
	instanceVariableNames: 'shutdown'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

!AJPPharoServer class methodsFor: 'accessing' stamp: 'pmm 2/25/2012 10:56'!
bufferSize
	"the buffer size to be used on sockets"
	^ 16r2000! !

!AJPPharoServer methodsFor: 'private' stamp: 'pmm 10/31/2010 14:17'!
acceptLoopOn: aSocket
	[ [ self shouldStop not and: [ aSocket isValid ] ] whileTrue: [
		| connection |
		connection := aSocket waitForAcceptFor: 3.
		connection isNil ifFalse: [
			self startWorkerOn: (self configureSocket: connection) ]
	 ] ] ensure: [
		aSocket destroy.
		shutdown signal ]! !

!AJPPharoServer methodsFor: 'private' stamp: 'pmm 2/9/2012 19:52'!
configureSocket: aSocket
	"http://baus.net/on-tcp_cork"
	^ aSocket
		setOption: 'TCP_NODELAY' value: 1;
		"setOption: 'TCP_CORK' value: 1;"
		"setOption: 'TCP_NOWAIT' value: 1;"
		"setOption: 'TCP_NOPUSH' value: 1;"
		"setOption: 'TCP_MAXSEG' value: AJPPharoServer bufferSize + (320 / 8);"
		setOption: 'SO_SNDBUF' value: AJPPharoServer bufferSize;
		setOption: 'SO_RCVBUF' value: AJPPharoServer bufferSize;
		yourself! !

!AJPPharoServer methodsFor: 'private' stamp: 'pmm 10/31/2010 14:16'!
newTcpSocket
	^ self configureSocket: Socket newTCP! !

!AJPPharoServer methodsFor: 'actions' stamp: 'pmm 10/30/2010 22:54'!
removeWorker: aLoop
	mutex critical: [
		workers remove: aLoop ifAbsent: [ "ignore" ] ] ! !

!AJPPharoServer methodsFor: 'actions' stamp: 'pmm 10/8/2010 11:34'!
startOn: aPort using: anAdapter
	| socket process |
	self beStarting.
	shutdown := Semaphore new.
	adapter := anAdapter.
	socket := self newTcpSocket.
	socket listenOn: aPort backlogSize: self backlogSize.
	process := [ self acceptLoopOn: socket ] newProcess.
	process priority: Processor highIOPriority.
	process name: 'AJP server'.
	process resume.
	self beStarted! !

!AJPPharoServer methodsFor: 'private' stamp: 'pmm 5/1/2012 18:22'!
startWorkerOn: aSocket
	| process |
	mutex critical: [
		| worker |
		shouldStop ifTrue: [
			aSocket closeAndDestroy: 30.
			^ self ].
		worker := AJPPharoSocketLoop on: aSocket server: self adapter: adapter.
		process := [ worker loop ] newProcess.
		process priority: Processor highIOPriority.
		process name: 'AJP worker ', (workers size  + 1) printString.
		workers add: worker ].
	process resume! !

!AJPPharoServer methodsFor: 'actions' stamp: 'pmm 10/8/2010 11:36'!
stop
	super stop.
	shutdown wait! !

AJPDefaultDataStream subclass: #AJPPharoSocketDataStream
	instanceVariableNames: 'socket readStream writeStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

!AJPPharoSocketDataStream class methodsFor: 'instance creation' stamp: 'pmm 2/25/2012 10:55'!
on: aSocket
	^ self basicNew initializeOn: aSocket! !

!AJPPharoSocketDataStream methodsFor: 'public' stamp: 'pmm 9/13/2013 11:05'!
destroy
	socket ifNotNil: [ :s |
		"#destroy may have been called several times concurrently"
		s closeAndDestroy: 30 ].
	socket := nil.
	readStream := nil.
	writeStream := nil
	! !

!AJPPharoSocketDataStream methodsFor: 'public' stamp: 'pmm 10/30/2010 19:53'!
flush
	writeStream flush! !

!AJPPharoSocketDataStream methodsFor: 'hinting' stamp: 'pmm 10/30/2010 18:32'!
hintNewBinaryPacket
	readStream readPacket! !

!AJPPharoSocketDataStream methodsFor: 'hinting' stamp: 'pmm 10/30/2010 18:32'!
hintNewTextPacket
	readStream readPacket! !

!AJPPharoSocketDataStream methodsFor: 'initialize-release' stamp: 'pmm 10/30/2010 19:08'!
initializeOn: aSocket
	self initialize.
	socket := aSocket.
	readStream := AJPPharoSocketReadStream on: aSocket.
	writeStream := AJPPharoSocketWriteStream on: aSocket! !

!AJPPharoSocketDataStream methodsFor: 'testing' stamp: 'pmm 2/25/2012 10:49'!
isConnected
	"The stream is connected if the socket is."

	^ socket notNil and: [ socket isConnected ]! !

!AJPPharoSocketDataStream methodsFor: 'reading' stamp: 'pmm 10/30/2010 18:32'!
readByte
	^ readStream nextByte! !

!AJPPharoSocketDataStream methodsFor: 'reading' stamp: 'pmm 10/30/2010 22:03'!
readChunk: anInteger into: aCollection startingAt: startIndex
	readStream readChunk: anInteger into: aCollection startingAt: startIndex! !

!AJPPharoSocketDataStream methodsFor: 'reading' stamp: 'pmm 2/25/2012 11:04'!
readStringWithSizeHighByte: aByte
	"reads a String with the high byte of the size already read
	otherwise the same as #readString"
	| result size |
	size := (aByte << 8) bitOr: (self readByte).
	size = 16rFFFF ifTrue: [
		^ String new ].
	
	"convert bytes to Characters, we don't want to switch to text mode because that would reset the buffers"
	result := readStream nextCharacters: size.
	
	"skip 0 byte"
	readStream nextByte = 0
		ifFalse: [ self protocolError: '0x00 byte string terminator expected' ].
	^ result! !

!AJPPharoSocketDataStream methodsFor: 'writing' stamp: 'pmm 10/30/2010 18:32'!
writeByte: aByte
	writeStream writeByte: aByte! !

!AJPPharoSocketDataStream methodsFor: 'writing' stamp: 'pmm 10/30/2010 18:32'!
writeBytes: aByteArray
	writeStream writeBytes: aByteArray! !

!AJPPharoSocketDataStream methodsFor: 'writing' stamp: 'pmm 12/26/2010 10:09'!
writeChunk: aBuffer
	writeStream writeChunk: aBuffer! !

!AJPPharoSocketDataStream methodsFor: 'writing' stamp: 'pmm 10/30/2010 18:32'!
writeString: aString
	self writeInteger: aString size.
	writeStream writeString:  aString.
	self writeByte: 0! !

AJPAdapterLoop subclass: #AJPPharoSocketLoop
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

!AJPPharoSocketLoop methodsFor: 'private' stamp: 'pmm 4/8/2012 12:28'!
createBufferDataStreamOn: aSocket
	^ AJPPharoSocketDataStream on: aSocket! !

!AJPPharoSocketLoop methodsFor: 'initialize-release' stamp: 'pmm 10/31/2010 13:44'!
initializeOn: aSocket server: aServer
	super
		initializeOn: (self createBufferDataStreamOn: aSocket)
		server: aServer! !

!AJPPharoSocketLoop methodsFor: 'reading' stamp: 'pmm 4/8/2012 23:11'!
loop
	[ [ self loopPossiblyThrowingExceptions ] repeat ]
		on: ConnectionClosed, AJPLoopTerminatedError, Error
		do: [ :error |
			stream destroy.
			server removeWorker: self.
			^ self ]! !

!AJPPharoSocketLoop methodsFor: 'private' stamp: 'pmm 2/25/2012 10:54'!
readIgnoringTimeout
	[ 
		| next |
		(shouldStop or: [ (stream isConnected) not ])
			ifTrue: [ AJPLoopTerminatedError signal ].
		next := [ stream readByte ]
			on: ConnectionTimedOut
			do: [ :error | nil ].
		next isNil ifFalse: [ ^ next ] ] repeat! !

GRError subclass: #AJPPharoSocketError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

GRObject subclass: #AJPPharoSocketStream
	instanceVariableNames: 'socket buffer position'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

AJPPharoSocketStream subclass: #AJPPharoSocketReadStream
	instanceVariableNames: 'limit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

!AJPPharoSocketReadStream methodsFor: 'private' stamp: 'pmm 2/25/2012 10:56'!
capacity
	"4 - 1 + 1"
	^ limit - position + 1! !

!AJPPharoSocketReadStream methodsFor: 'private' stamp: 'pmm 4/8/2012 12:48'!
ensureCapacity: anInteger
	| capacity |
	capacity := self capacity.
	capacity < anInteger ifTrue: [
		anInteger >  16r2000 "AJPPharoServer bufferSize" ifTrue: [
			self socketError: 'data size ', anInteger printString, ' not supported' ].
		(position + anInteger - 1) <= 16r2000 "AJPPharoServer bufferSize" 
			ifTrue: [
				self readntoBuffer: anInteger - capacity ]
			ifFalse: [
				"uncool"
				buffer copyFrom: position to: limit into: buffer startingAt: 1.
				limit := limit - position - 1.
				position := 1 ] ]! !

!AJPPharoSocketReadStream methodsFor: 'initialization' stamp: 'pmm 10/30/2010 19:44'!
initialize
	super initialize.
	limit := 0! !

!AJPPharoSocketReadStream methodsFor: 'reading' stamp: 'pmm 10/30/2010 18:18'!
nextByte
	| data |
	self ensureCapacity: 1.
	data := buffer at: position.
	position := position + 1.
	^ data! !

!AJPPharoSocketReadStream methodsFor: 'reading' stamp: 'pmm 10/30/2010 18:18'!
nextBytes: anInteger
	| data |
	self ensureCapacity: anInteger.
	data := buffer copyFrom: position to: position + anInteger.
	position := position + anInteger.
	^ data! !

!AJPPharoSocketReadStream methodsFor: 'reading' stamp: 'pmm 4/8/2012 12:33'!
nextCharacters: anInteger
	| data |
	self ensureCapacity: anInteger.
	data := ByteString new: anInteger.
	data replaceFrom: 1 to: anInteger with: buffer startingAt: position.
	position := position + anInteger.
	^ data! !

!AJPPharoSocketReadStream methodsFor: 'reading' stamp: 'pmm 10/30/2010 22:05'!
readChunk: anInteger into: aCollection startingAt: startIndex
	self ensureCapacity: anInteger.
	aCollection replaceFrom: startIndex to: startIndex + anInteger - 1 with: buffer startingAt: position.
	position := position + anInteger! !

!AJPPharoSocketReadStream methodsFor: 'reading' stamp: 'pmm 10/30/2010 22:36'!
readPacket
	position <= limit
		ifTrue: [
			"still data available, copy to start of buffer for further reads"
			buffer := buffer replaceFrom: 1 to: limit - position + 1 with: buffer startingAt: position.
			limit := limit - position + 1.
			position := 1 ]
		ifFalse: [
			| read |
			read := socket receiveAvailableDataInto: buffer startingAt: 1.
			[ read = 0 ] whileTrue: [
				socket waitForData.
				read := socket receiveAvailableDataInto: buffer startingAt: 1 ].
			position := 1.
			limit := read ]! !

!AJPPharoSocketReadStream methodsFor: 'private' stamp: 'pmm 4/8/2012 12:37'!
readntoBuffer: anInteger
	| read |
	read := socket receiveAvailableDataInto: buffer startingAt: limit + 1.
	limit := limit + read.
	read < anInteger
		ifTrue: [
			socket waitForData.
			self readntoBuffer: anInteger - read ]! !

!AJPPharoSocketStream class methodsFor: 'instance creation' stamp: 'pmm 2/25/2012 10:56'!
on: aSocket
	^ self basicNew initializeOn: aSocket! !

!AJPPharoSocketStream methodsFor: 'initialize-release' stamp: 'pmm 10/30/2010 18:26'!
initializeOn: aSocket
	self initialize.
	socket := aSocket.
	position := 1.
	buffer := self newBuffer! !

!AJPPharoSocketStream methodsFor: 'private' stamp: 'pmm 2/25/2012 10:56'!
newBuffer
	^ ByteArray new: AJPPharoServer bufferSize withAll: 0! !

!AJPPharoSocketStream methodsFor: 'private' stamp: 'pmm 4/23/2011 15:47'!
socketError: aString
	AJPPharoSocketError signal: aString! !

AJPPharoSocketStream subclass: #AJPPharoSocketWriteStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AJP-Pharo-Core'!

!AJPPharoSocketWriteStream methodsFor: 'public' stamp: 'pmm 10/30/2010 18:19'!
flush
	self sendBuffer! !

!AJPPharoSocketWriteStream methodsFor: 'private' stamp: 'pmm 10/30/2010 22:27'!
sendBuffer
	socket sendData: buffer count: position - 1.
	position := 1! !

!AJPPharoSocketWriteStream methodsFor: 'writing' stamp: 'pmm 9/13/2013 11:05'!
writeByte: aByte
	buffer at: position put: aByte.
	position := position + 1! !

!AJPPharoSocketWriteStream methodsFor: 'writing' stamp: 'pmm 4/7/2012 14:54'!
writeBytes: aByteArray
	| arraySize |
	arraySize := aByteArray size.
	buffer replaceFrom: position to: position + arraySize - 1 with: aByteArray startingAt: 1.
	position := position + arraySize! !

!AJPPharoSocketWriteStream methodsFor: 'writing' stamp: 'pmm 4/6/2012 14:10'!
writeChunk: aBuffer
	aBuffer sendToSocket: socket! !

!AJPPharoSocketWriteStream methodsFor: 'writing' stamp: 'pmm 9/13/2013 11:05'!
writeString: aString
	| stringSize |
	stringSize := aString size.
	buffer replaceFrom: position to: position + stringSize - 1 with: aString startingAt: 1.
	position := position + stringSize! !
